import tabulate
import typer
import yaml

from munch import Munch
from os import listdir
from os.path import isfile, join


app = typer.Typer(add_completion=False)


@app.command()
def main(
    images: str = typer.Option('etc/images/', help='Path to the directory containing all image files')
):
    CONF = Munch.fromDict(locals())

    onlyfiles = []
    for f in listdir(CONF.images):
        if isfile(join(CONF.images, f)):
            onlyfiles.append(f)

    all_images = []
    for file in onlyfiles:
        with open(join(CONF.images, file)) as fp:
            data = yaml.load(fp, Loader=yaml.SafeLoader)
            images = data.get('images')
            for image in images:
                all_images.append(image)

    data = []
    for image in all_images:
        data.append([image['name'], image['login'], image.get('password', '')])

    result = tabulate.tabulate(
        sorted(data),
        headers=['Name', 'Login user', 'Password'],
        tablefmt="rst"
    )
    print(result)


if __name__ == '__main__':
    app()



from loguru import logger
from munch import Munch
from unittest import TestCase, mock
from openstack.image.v2.image import Image
from openstack.image.v2._proxy import Proxy

from src import manage

logger.remove()   # disable all logging from manage.py

# sample config from images.yml
FAKE_YML = '''
---
images:
  - name: Ubuntu 20.04
    format: qcow2
    login: ubuntu
    min_disk: 8
    min_ram: 512
    status: active
    visibility: public
    multi: true
    meta:
      os_distro: ubuntu
      os_version: '20.04'
    tags: []
    versions:
      - version: '1'
        url: http://url.com
        checksum: '1234'
'''

# sample image dict as generated from FAKE_YML
FAKE_IMAGE_DICT = {
    'name': 'Ubuntu 20.04',
    'format': 'qcow2',
    'login': 'ubuntu',
    'min_disk': 8,
    'min_ram': 512,
    'status': 'active',
    'visibility': 'public',
    'multi': True,
    'meta': {
        'os_distro': 'ubuntu',
        'os_version': '20.04'
    },
    'tags': [],
    'versions': [
        {
            'version': '1',
            'url': 'http://url.com',
            'checksum': '1234'
        }
    ]
}

# data to generate a fake openstack.image.v2.image.Image object
FAKE_IMAGE_DATA = {
    'id': '123456789abcdef',
    'name': FAKE_IMAGE_DICT['name'],
    'container_format': 'bare',
    'disk_format': FAKE_IMAGE_DICT['format'],
    'min_disk': FAKE_IMAGE_DICT.get('min_disk', 0),
    'min_ram': FAKE_IMAGE_DICT.get('min_ram', 0),
    'size': 123456789,
    'protected': False,
    'status': 'active',
    'tags': ['fake_tag'],
    'os_hidden': False,
    'visibility': 'public',
    'os_distro': FAKE_IMAGE_DICT['meta']['os_distro'],
    'os_version': FAKE_IMAGE_DICT['meta']['os_version'],
    'properties': {
        'image_original_user': FAKE_IMAGE_DICT['login'],
        'internal_version': FAKE_IMAGE_DICT['versions'][0]['version']
    }
}


class TestManage(TestCase):

    def setUp(self):
        ''' create all necessary test data, gets called before each test '''

        self.fake_image_dict = FAKE_IMAGE_DICT.copy()
        self.fake_image = Image(**FAKE_IMAGE_DATA)
        self.fake_name = '%s (%s)' % (self.fake_image_dict['name'], '1')
        self.fake_url = 'http://url.com'
        self.versions = {'1': {'url': self.fake_url}}
        self.sorted_versions = ['2', '1']
        self.previous_image = self.fake_image
        self.imported_image = self.fake_image

        self.sot = manage.ImageManager()
        # since oslo_config.cfg.ConfigOpts objects allow attribute-style access,
        # we can mimick its behaviour with a munch.Munch object
        self.sot.CONF = Munch(
            latest=True,
            dry_run=False,
            use_os_hidden=False,
            delete=False,
            yes_i_really_know_what_i_do=False,
            hide=False,
            deactivate=False,
            cloud='fake-cloud',
            images='etc/images/',
            name=None,
            tag='fake_tag',
            filter='',
            check=False,
            validate=False
        )

        # we can also mimick an openstack connection object with a Munch
        self.sot.conn = Munch(
            current_project_id='123456789',
            image=Proxy
        )

    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.images')
    def test_get_images(self, mock_images):
        ''' test manage.ImageManager.get_images() '''

        mock_images.return_value = [self.fake_image]
        expected_result = {self.fake_image.name: self.fake_image}

        result = self.sot.get_images()
        mock_images.assert_called_once()
        self.assertEqual(result, expected_result)

        mock_images.reset_mock()

        # test with use_os_hidden = True
        self.sot.CONF.use_os_hidden = True
        result = self.sot.get_images()
        mock_images.assert_called_with(**{'os_hidden': True})
        self.assertEqual(mock_images.call_count, 2)
        self.assertEqual(result, expected_result)

    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.get_image')
    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.import_image')
    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.create_image')
    def test_import_image(self, mock_create, mock_import, mock_get_image):
        ''' test manage.ImageManager.import_image() '''

        mock_create.return_value = self.fake_image
        mock_get_image.return_value = self.fake_image
        properties = {
            'container_format': 'bare',
            'disk_format': self.fake_image_dict['format'],
            'min_disk': self.fake_image_dict.get('min_disk', 0),
            'min_ram': self.fake_image_dict.get('min_ram', 0),
            'name': self.fake_name,
            'tags': [self.sot.CONF.tag],
            'visibility': 'private'
        }

        self.sot.import_image(self.fake_image_dict, self.fake_name, self.fake_url)

        mock_create.assert_called_once_with(**properties)
        mock_import.assert_called_once_with(self.fake_image, method='web-download', uri=self.fake_url)
        mock_get_image.assert_called_once_with(self.fake_image)

    @mock.patch('src.manage.ImageManager.set_properties')
    @mock.patch('src.manage.ImageManager.import_image')
    @mock.patch('src.manage.requests.head')
    @mock.patch('src.manage.ImageManager.get_images')
    def test_process_image(self, mock_get_images, mock_requests, mock_import_image, mock_set_properties):
        ''' test manage.ImageManager.process_image() '''

        mock_requests.return_value.status_code = 200
        meta = self.fake_image_dict['meta']

        result = self.sot.process_image(self.fake_image_dict, self.versions, self.sorted_versions, meta)

        self.assertEqual(mock_get_images.call_count, 2)
        mock_requests.assert_called_once_with(self.fake_url)
        mock_import_image.assert_called_once_with(self.fake_image_dict, self.fake_name, self.fake_url)
        mock_set_properties.assert_called_once_with(self.fake_image_dict, self.fake_name, self.versions, '1', '', meta)
        self.assertEqual(result, ({self.fake_image_dict['name']}, mock_get_images.return_value.__getitem__(), None))

        mock_get_images.reset_mock()
        mock_requests.reset_mock()
        mock_import_image.reset_mock()
        mock_set_properties.reset_mock()

        # test the same function with dry_run = True
        self.sot.CONF.dry_run = True
        result = self.sot.process_image(self.fake_image_dict, self.versions, self.sorted_versions, meta)

        mock_get_images.assert_called_once()
        mock_requests.assert_called_once_with(self.fake_url)
        mock_import_image.assert_not_called()
        mock_set_properties.assert_not_called()
        self.assertEqual(result, ({self.fake_image_dict['name']}, None, None))

    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.deactivate_image')
    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.remove_tag')
    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.add_tag')
    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.update_image')
    @mock.patch('src.manage.ImageManager.get_images')
    def test_set_properties(self, mock_get_images, mock_update_image, mock_add_tag, mock_remove_tag, mock_deactivate):
        ''' test manage.ImageManager.set_properties() '''

        meta = self.fake_image_dict['meta']
        mock_get_images.return_value = {self.fake_name: self.fake_image}

        self.fake_image_dict['tags'] = ['my_tag']
        self.fake_image_dict['status'] = 'deactivated'

        self.sot.set_properties(self.fake_image_dict, self.fake_name, self.versions, '1', '', meta)

        mock_get_images.assert_called_once()
        mock_update_image.assert_called()
        mock_add_tag.assert_called_once_with(self.fake_image.id, 'my_tag')
        mock_remove_tag.assert_called_once_with(self.fake_image.id, 'fake_tag')
        mock_deactivate.assert_called_once_with(self.fake_image.id)

    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.update_image')
    @mock.patch('src.manage.ImageManager.get_images')
    def test_rename_images(self, mock_get_images, mock_update_image):
        ''' test manage.ImageManager.rename_images() '''

        # test with len(sorted_versions) > 1
        mock_get_images.return_value = {
            self.fake_name: self.fake_image,
            self.fake_image.name: self.fake_image
        }
        self.sot.rename_images(self.fake_image.name, self.sorted_versions, self.imported_image, self.previous_image)

        mock_get_images.assert_called_once()
        mock_update_image.assert_called_with(self.fake_image.id, name=mock.ANY)

        self.assertEqual(mock_update_image.call_count, 2)
        mock_get_images.reset_mock()
        mock_update_image.reset_mock()

        # test with len(sorted_versions) == 1 and name in cloud_images
        mock_get_images.return_value = {self.fake_image.name: self.fake_image}
        self.sorted_versions = ['1']

        self.sot.rename_images(self.fake_image.name, self.sorted_versions, self.imported_image, self.previous_image)
        mock_get_images.assert_called_once()
        mock_update_image.assert_called_with(self.fake_image.id, name=mock.ANY)
        self.assertEqual(mock_update_image.call_count, 2)

        mock_get_images.reset_mock()
        mock_update_image.reset_mock()

        # test with len(sorted_versions) == 1
        mock_get_images.return_value = {self.fake_name: self.fake_image}

        self.sot.rename_images(self.fake_image.name, self.sorted_versions, self.imported_image, self.previous_image)
        mock_get_images.assert_called_once()
        mock_update_image.assert_called_once_with(self.fake_image.id, name=mock.ANY)

    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.delete_image')
    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.update_image')
    @mock.patch('src.manage.openstack.image.v2._proxy.Proxy.deactivate_image')
    @mock.patch('src.manage.ImageManager.get_images')
    def test_manage_outdated_images(self, mock_get_images, mock_deactivate, mock_update_image, mock_delete_image):
        ''' test manage.ImageManager.manage_outdated_images '''

        # test deletion of images
        self.sot.CONF.delete = True
        self.sot.CONF.yes_i_really_know_what_i_do = True
        managed_images = {'some_image_name'}
        mock_get_images.return_value = {
            self.fake_image.name: self.fake_image
        }

        self.sot.manage_outdated_images(managed_images)
        mock_get_images.assert_called_once()
        mock_deactivate.assert_called_once_with(self.fake_image.id)
        mock_update_image.assert_called_once_with(self.fake_image.id, visibility='community')
        mock_delete_image.assert_called_once_with(self.fake_image.id)

        mock_get_images.reset_mock()
        mock_deactivate.reset_mock()
        mock_update_image.reset_mock()
        mock_delete_image.reset_mock()

        # test hide and deactivate of images
        self.sot.CONF.delete = False
        self.sot.CONF.yes_i_really_know_what_i_do = False
        self.sot.CONF.hide = True
        self.sot.CONF.deactivate = True

        self.sot.manage_outdated_images(managed_images)
        mock_get_images.assert_called_once()
        mock_deactivate.assert_called_once_with(self.fake_image.id)
        mock_update_image.assert_called_once_with(self.fake_image.id, visibility='community')
        mock_delete_image.assert_not_called()

    @mock.patch('src.manage.ImageManager.validate_yaml_schema')
    @mock.patch('src.manage.ImageManager.check_image_metadata')
    @mock.patch('src.manage.ImageManager.manage_outdated_images')
    @mock.patch('src.manage.ImageManager.process_images')
    @mock.patch('src.manage.ImageManager.get_images')
    @mock.patch('src.manage.ImageManager.read_image_files')
    @mock.patch('src.manage.openstack.connect')
    def test_main(self, mock_connect, mock_read_image_files, mock_get_images,
                  mock_process_images, mock_manage_outdated, mock_check_metadata, mock_validate_yaml):
        ''' test manage.ImageManager.main() '''
        mock_read_image_files.return_value = [self.fake_image_dict]

        self.sot.main()

        mock_connect.assert_called_once_with(cloud=self.sot.CONF.cloud)
        mock_read_image_files.assert_called_once()
        mock_get_images.assert_called_once()
        mock_process_images.assert_called_once_with([self.fake_image_dict])
        mock_manage_outdated.assert_called_once_with(set())
        mock_check_metadata.assert_not_called()
        mock_validate_yaml.assert_not_called()

        # reset
        mock_read_image_files.reset_mock()
        mock_get_images.reset_mock()
        mock_process_images.reset_mock()
        mock_manage_outdated.reset_mock()
        mock_check_metadata.reset_mock()
        mock_validate_yaml.reset_mock()

        # test with dry_run = True and validate = True
        self.sot.CONF.dry_run = True
        self.sot.CONF.validate = True

        self.sot.main()
        mock_read_image_files.assert_not_called()
        mock_get_images.assert_not_called()
        mock_process_images.assert_not_called()
        mock_manage_outdated.assert_not_called()
        mock_check_metadata.assert_called_once()
        mock_validate_yaml.assert_not_called()

        # reset
        mock_read_image_files.reset_mock()
        mock_get_images.reset_mock()
        mock_process_images.reset_mock()
        mock_manage_outdated.reset_mock()
        mock_check_metadata.reset_mock()
        mock_validate_yaml.reset_mock()

        # test with check = True
        self.sot.CONF.check = True

        self.sot.CONF.dry_run = False
        self.sot.CONF.validate = False

        self.sot.main()
        mock_read_image_files.assert_not_called()
        mock_get_images.assert_not_called()
        mock_process_images.assert_not_called()
        mock_manage_outdated.assert_not_called()
        mock_check_metadata.assert_not_called()
        mock_validate_yaml.assert_called_once()

    @mock.patch('src.manage.os.path.isfile')
    @mock.patch('src.manage.os.listdir')
    @mock.patch('builtins.open', mock.mock_open(read_data=str(FAKE_YML)))
    def test_read_image_files(self, mock_listdir, mock_isfile):
        ''' test manage.ImageManager.read_image_files() '''
        mock_listdir.return_value = ['fake.yml']
        mock_isfile.return_value = True

        result = self.sot.read_image_files()
        self.assertEqual(result, [FAKE_IMAGE_DICT])

    @mock.patch('src.manage.ImageManager.rename_images')
    @mock.patch('src.manage.ImageManager.process_image')
    def test_process_images(self, mock_process_image, mock_rename_images):
        ''' test manage.ImageManager.process_images() '''
        meta = self.fake_image_dict['meta']
        self.fake_image_dict['tags'] = [self.sot.CONF.tag, 'os:%s' % self.fake_image_dict['meta']['os_distro']]
        mock_process_image.return_value = ({self.fake_image_dict['name']}, self.imported_image, self.previous_image)

        result = self.sot.process_images([self.fake_image_dict])

        mock_process_image.assert_called_once_with(self.fake_image_dict, self.versions, ['1'], meta)
        mock_rename_images.assert_called_once_with(self.fake_image_dict['name'], ['1'],
                                                   self.imported_image,
                                                   self.previous_image)

        self.assertEqual(result, {self.fake_image_dict["name"]})



import time
import openstack
import requests
import yaml
import os
import re
import sys
import typer
import yamale

from datetime import datetime
from decimal import Decimal, ROUND_UP
from loguru import logger
from munch import Munch
from natsort import natsorted
from typing import List, Optional
from yamale import YamaleError
from openstack.image.v2.image import Image
from openstack.exceptions import DuplicateResource


class ImageManager:

    def __init__(self) -> None:
        self.exit_with_error = False

    def create_cli_args(
        self,
        debug: bool = typer.Option(False, '--debug', help='Enable debug logging'),
        dry_run: bool = typer.Option(False, '--dry-run', help='Do not perform any changes'),
        latest: bool = typer.Option(False, '--latest', help='Only import the latest version for images of type multi'),
        cloud: str = typer.Option('openstack', help='Cloud name in clouds.yaml'),
        images: str = typer.Option('etc/images/', help='Path to the directory containing all image files or path to '
                                   'a single image file'),
        name: Optional[List[str]] = typer.Option([], help='Name of the image to process, '
                                                 'use repeatedly for multiple images'),
        tag: str = typer.Option('managed_by_osism', help='Name of the tag used to identify managed images'),
        filter: str = typer.Option(None, help='Filter images with a regex on their name'),
        deactivate: bool = typer.Option(False, '--deactivate', help='Deactivate images that should be deleted'),
        hide: bool = typer.Option(False, '--hide', help='Hide images that should be deleted'),
        delete: bool = typer.Option(False, '--delete', help='Delete outdated images'),
        yes_i_really_know_what_i_do: bool = typer.Option(False, '--yes-i-really-know-what-i-do',
                                                         help='Really delete images'),
        use_os_hidden: bool = typer.Option(False, '--use-os-hidden', help='Use the os_hidden property'),
        check: bool = typer.Option(False, '--check',
                                   help='Check the local image definitions against the SCS Image Metadata Standard'),
        validate: bool = typer.Option(False, '--validate',
                                      help='Validate the image metadata against the SCS Image Metadata Standard')
    ):
        self.CONF = Munch.fromDict(locals())
        self.CONF.pop('self')   # remove the self object from CONF

        if self.CONF.debug:
            level = "DEBUG"
        else:
            level = "INFO"

        logger.remove()    # remove the default sink
        log_fmt = ("<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | "
                   "<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>")
        logger.add(sys.stderr, format=log_fmt, level=level, colorize=True)

        if __name__ == '__main__':
            self.main()

    def read_image_files(self) -> list:
        """ Read all YAML files in self.CONF.images """
        image_files = []

        if os.path.isdir(self.CONF.images):
            for file in os.listdir(self.CONF.images):
                if file.endswith(tuple([".yml", "yaml"])):
                    image_files.append(file)
        else:
            image_files = [self.CONF.images]

        all_images = []
        for file in image_files:
            if os.path.isdir(self.CONF.images):
                file = os.path.join(self.CONF.images, file)
            with open(file) as fp:
                try:
                    data = yaml.load(fp, Loader=yaml.SafeLoader)
                    images = data.get('images')
                    for image in images:
                        if self.CONF.filter:
                            if re.search(self.CONF.filter, image["name"]):
                                all_images.append(image)
                        else:
                            all_images.append(image)
                except yaml.YAMLError as exc:
                    logger.error(exc)
        return all_images

    def get_checksum(self, url: str, checksums_url: str) -> str:
        '''
        Get the checksum of an upstream image by parsing its corresponding checksums file

        Params:
            url: the download URL of the image
            checksums_url: the URL of the corresponding checksums file

        Returns:
            the matching checksum, if it is available or else an empty string
        '''
        filename = url.split('/')[-1]
        checksums_file = requests.get(checksums_url).text
        for line in checksums_file.splitlines():
            if filename in line:
                split = line.split(' ')
                for elem in split:
                    if (len(elem) == 128 or len(elem) == 64 or len(elem) == 40 or len(elem) == 32) and '.' not in elem:
                        return elem
        return ''

    def create_connection(self) -> None:
        if "OS_AUTH_URL" in os.environ:
            self.conn = openstack.connect()
        else:
            self.conn = openstack.connect(cloud=self.CONF.cloud)

    def main(self) -> None:
        '''
        Read all files in etc/images/ and process each image
        Rename outdated images when not dry-running
        '''
        logger.debug("cloud = %s" % self.CONF.cloud)
        logger.debug("dry-run = %s" % self.CONF.dry_run)
        logger.debug("images = %s" % self.CONF.images)
        logger.debug("tag = %s" % self.CONF.tag)
        logger.debug("yes-i-really-know-what-i-do = %s" % self.CONF.yes_i_really_know_what_i_do)

        # validate metadata of existing images
        if self.CONF.validate:
            self.create_connection()
            self.check_image_metadata()

        # check local image definitions with yamale
        elif self.CONF.check:
            self.validate_yaml_schema()

        # manage images
        else:
            self.create_connection()
            images = self.read_image_files()
            managed_images = set()

            # get all active managed images, so they don't get deleted when using --name
            cloud_images = self.get_images()
            for image in cloud_images:
                if cloud_images[image].visibility == "public":
                    managed_images.add(image)

            managed_images = set.union(managed_images, self.process_images(images))

            if not self.CONF.dry_run:
                self.manage_outdated_images(managed_images)

        if self.exit_with_error:
            sys.exit('\nERROR: One or more errors occurred during the execution of the program, '
                     'please check the output.')

    def process_images(self, images) -> set:
        """ Process each image from images.yaml """

        REQUIRED_KEYS = [
            'format',
            'name',
            'login',
            'status',
            'versions',
            'visibility',
        ]

        for image in images:

            for required_key in REQUIRED_KEYS:
                if required_key not in image:
                    logger.error("'%s' lacks the necessary key %s" % (image['name'], required_key))
                    self.exit_with_error = True
                    continue

            if self.CONF.name and image['name'] not in self.CONF.name:
                continue

            logger.debug("Processing '%s'" % image['name'])

            try:
                versions = dict()
                for version in image['versions']:
                    versions[str(version['version'])] = {
                        'url': version['url']
                    }
                    if 'visibility' in version:
                        versions[version['version']]['visibility'] = version['visibility']
                    if 'os_version' in version:
                        versions[version['version']]['os_version'] = version['os_version']
                    if 'hidden' in version:
                        versions[version['version']]['hidden'] = version['hidden']
                    if version['version'] == 'latest':
                        if 'checksums_url' in version:
                            versions[version['version']]['checksums_url'] = version['checksums_url']
                        else:
                            raise Exception()
                    if 'meta' in version:
                        versions[version['version']]['meta'] = version['meta']
            except Exception:
                logger.error('Key "checksums_url" is required when using version "latest"')
                continue

            sorted_versions = natsorted(versions.keys())
            image['tags'].append(self.CONF.tag)

            if 'os_distro' in image['meta']:
                image['tags'].append("os:%s" % image['meta']['os_distro'])

            existing_images, imported_image, previous_image = \
                self.process_image(image, versions, sorted_versions, image['meta'].copy())

            if imported_image and image['multi']:
                self.rename_images(image['name'], sorted_versions, imported_image, previous_image)

        return existing_images

    def import_image(self, image: dict, name: str, url: str) -> Image:
        '''
        Create a new image in Glance and upload it using the web-download method

        Params:
            image: image dict from images.yml
            name: name of the image to import
            url: download URL of the image
        '''
        logger.info("Importing image %s" % name)

        properties = {
            'container_format': 'bare',
            'disk_format': image['format'],
            'min_disk': image.get('min_disk', 0),
            'min_ram': image.get('min_ram', 0),
            'name': name,
            'tags': [self.CONF.tag],
            'visibility': 'private'
        }

        new_image = self.conn.image.create_image(**properties)
        self.conn.image.import_image(new_image, method='web-download', uri=url)

        while True:
            try:
                imported_image = self.conn.image.get_image(new_image)
                if imported_image.status != 'active':
                    logger.info("Waiting for import to complete...")
                    time.sleep(10.0)
                else:
                    break
            except Exception as e:
                logger.error("Exception while importing image %s\n%s" % (name, e))
                self.exit_with_error = True
        return imported_image

    def get_images(self) -> dict:
        '''
        Load all images from OpenStack and filter by the tag set by the --tag CLI option

        Returns:
            a dict containing all matching images as openstack.image.v2.image.Image objects
        '''
        result = {}

        for image in self.conn.image.images():
            if self.CONF.tag in image.tags and (image.visibility == 'public'
                                                or image.owner == self.conn.current_project_id):
                result[image.name] = image
                logger.debug("Managed image '%s' (tags = %s)" % (image.name, image.tags))
            else:
                logger.debug("Unmanaged image '%s' (tags = %s)" % (image.name, image.tags))

        if self.CONF.use_os_hidden:
            for image in self.conn.image.images(**{'os_hidden': True}):
                if self.CONF.tag in image.tags and (image.visibility == 'public'
                                                    or image.owner == self.conn.current_project_id):
                    result[image.name] = image
                    logger.debug("Managed hidden image '%s' (tags = %s)" % (image.name, image.tags))
                else:
                    logger.debug("Unmanaged hidden image '%s' (tags = %s)" % (image.name, image.tags))
        return result

    def process_image(self, image: dict, versions: dict, sorted_versions: list, meta: dict) -> tuple:
        '''
        Process one image from etc/images/
        Check if the image already exists in Glance and import it if not

        Params:
            image: image dict from images.yml
            versions: versions dict generated by main()
            sorted_versions: list with all sorted image versions
            meta: metadata of the image, does not include version-specific metadata

        Returns:
            Tuple with (existing_images, imported_image, previous_image)
        '''
        cloud_images = self.get_images()

        existing_images = set()
        imported_image = None
        previous_image = None
        upstream_checksum = ''

        for version in sorted_versions:
            if image['multi']:
                name = "%s (%s)" % (image['name'], version)
            else:
                name = "%s %s" % (image['name'], version)

            logger.info("Processing image '%s'" % name)
            logger.debug("Checking existence of '%s'" % name)
            existence = name in cloud_images

            if image['multi'] and self.CONF.latest and version == sorted_versions[-1] and not existence:
                existence = image['name'] in cloud_images
                try:
                    if existence:
                        existence = cloud_images[image['name']]['properties']['internal_version'] == version
                except KeyError:
                    logger.error("Image %s is missing property 'internal_version'" % image['name'])

            elif (image['multi'] and len(sorted_versions) > 1 and version == sorted_versions[-1]
                  and not existence):
                previous = "%s (%s)" % (image['name'], sorted_versions[-2])
                existence = previous in cloud_images and image['name'] in cloud_images

            elif (image['multi'] and len(sorted_versions) > 1 and version == sorted_versions[-2]
                  and not existence):
                existence = image['name'] in cloud_images

            elif image['multi'] and len(sorted_versions) == 1:
                existence = image['name'] in cloud_images

            if version == 'latest':
                checksums_url = versions[version]['checksums_url']
                upstream_checksum = self.get_checksum(versions[version]['url'], checksums_url)
                if not upstream_checksum:
                    logger.error("Could not find checksum for image '%s', check the checksums_url" % image['name'])
                    return existing_images, imported_image, previous_image

                try:
                    image_checksum = (cloud_images[image['name']]['properties']['upstream_checksum']
                                      if image['name'] in cloud_images else '')
                    if image_checksum == upstream_checksum:
                        logger.info("No new version for '%s'" % image['name'])
                        existing_images.add(image['name'])
                        return existing_images, imported_image, previous_image
                    else:
                        logger.info("New version for '%s'" % image['name'])
                        existence = False
                except KeyError:
                    # when switching from a release pointer to a latest pointer, the image has no checksum property
                    existence = False

            if not existence and not (self.CONF.latest and len(sorted_versions) > 1 and version != sorted_versions[-1]):

                url = versions[version]['url']
                r = requests.head(url)

                if r.status_code in [200, 302]:
                    logger.info("Tested URL %s: %s" % (url, r.status_code))
                else:
                    logger.error("Tested URL %s: %s" % (url, r.status_code))
                    logger.error("Skipping '%s' due to HTTP status code %s" % (name, r.status_code))
                    self.exit_with_error = True
                    return existing_images, imported_image, previous_image

                if image['multi'] and image['name'] in cloud_images:
                    self.previous_image = cloud_images[image['name']]

                if not self.CONF.dry_run:
                    self.import_image(image, name, url)
                    logger.info("Import of '%s' successfully completed, reloading images" % name)
                    cloud_images = self.get_images()
                    imported_image = cloud_images[name]

            elif self.CONF.latest and version != sorted_versions[-1]:
                logger.info("Skipping image '%s' (only importing the latest version from type multi)" % name)

            if image['multi']:
                existing_images.add(image['name'])
            else:
                existing_images.add(name)

            if imported_image:
                self.set_properties(image.copy(), name, versions, version, upstream_checksum, meta)
        return existing_images, imported_image, previous_image

    def set_properties(self, image: dict, name: str, versions: dict, version: str, upstream_checksum: str, meta: dict):
        '''
        Set image properties and tags based on the configuration from images.yml

        Params:
            image: image dict from images.yml
            name: name of the image including the version string
            versions: versions dict generated by main()
            version: currently processed version
            meta: metadata of the image, does not include version-specific metadata
        '''
        cloud_images = self.get_images()
        image['meta'] = meta.copy()

        if name in cloud_images:
            logger.info("Checking parameters of '%s'" % name)

            cloud_image = cloud_images[name]
            real_image_size = int(Decimal(cloud_image.size / 2**30).quantize(Decimal('1.'), rounding=ROUND_UP))

            if 'min_disk' in image and image['min_disk'] != cloud_image.min_disk:
                logger.info("Setting min_disk: %s != %s" % (image['min_disk'], cloud_image.min_disk))
                self.conn.image.update_image(cloud_image.id, **{'min_disk': int(image['min_disk'])})

            if ('min_disk' in image and real_image_size > image['min_disk']) or 'min_disk' not in image:
                logger.info("Setting min_disk = %d" % real_image_size)
                self.conn.image.update_image(cloud_image.id, **{'min_disk': real_image_size})

            if 'min_ram' in image and image['min_ram'] != cloud_image.min_ram:
                logger.info("Setting min_ram: %s != %s" % (image['min_ram'], cloud_image.min_ram))
                self.conn.image.update_image(cloud_image.id, **{'min_ram': int(image['min_ram'])})

            if 'build_date' in versions[version]:
                logger.info("Setting image_build_date = %s" % versions[version]['build_date'])
                image['meta']['image_build_date'] = versions[version]['build_date']

            if self.CONF.use_os_hidden:
                if 'hidden' in versions[version]:
                    logger.info("Setting os_hidden = %s" % versions[version]['hidden'])
                    self.conn.image.update_image(cloud_image.id, **{'os_hidden': versions[version]['hidden']})

                elif version != natsorted(versions.keys())[-1:]:
                    logger.info("Setting os_hidden = True")
                    self.conn.image.update_image(cloud_image.id, **{'os_hidden': True})

            if version == 'latest':
                try:
                    url = versions[version]['url']
                    modify_date = requests.head(url, allow_redirects=True).headers['Last-Modified']

                    date_format = '%a, %d %b %Y %H:%M:%S %Z'
                    modify_date = str(datetime.strptime(modify_date, date_format).date())
                    modify_date = modify_date.replace('-', '')

                    logger.info("Setting internal_version = %s" % modify_date)
                    image['meta']['internal_version'] = modify_date
                except Exception:
                    logger.error("Error when retrieving the modification date of image '%s'", image['name'])
                    logger.info("Setting internal_version = %s" % version)
                    image['meta']['internal_version'] = version
            else:
                logger.info("Setting internal_version = %s" % version)
                image['meta']['internal_version'] = version

            logger.info("Setting image_original_user = %s" % image['login'])
            image['meta']['image_original_user'] = image['login']

            if version == 'latest' and upstream_checksum:
                image['meta']['upstream_checksum'] = upstream_checksum

            if image['multi'] and 'os_version' in versions[version]:
                image['meta']['os_version'] = versions[version]['os_version']
            elif not image['multi']:
                image['meta']['os_version'] = version

            for tag in image['tags']:
                if tag not in cloud_image.tags:
                    logger.info("Adding tag %s" % (tag))
                    self.conn.image.add_tag(cloud_image.id, tag)

            for tag in cloud_image.tags:
                if tag not in image['tags']:
                    logger.info("Deleting tag %s" % (tag))
                    self.conn.image.remove_tag(cloud_image.id, tag)

            if 'meta' in versions[version]:
                for key in versions[version]['meta'].keys():
                    image['meta'][key] = versions[version]['meta'][key]

            properties = cloud_image.properties
            for property in properties:
                if property in image['meta']:
                    if image['meta'][property] != properties[property]:
                        logger.info("Setting property %s: %s != %s" %
                                    (property, properties[property], image['meta'][property]))
                        self.conn.image.update_image(cloud_image.id, **{property: str(image['meta'][property])})

                elif property not in ['self', 'schema', 'stores'] or not property.startswith('os_'):
                    # FIXME: handle deletion of properties
                    logger.debug("Deleting property %s" % (property))

            for property in image['meta']:
                if property not in properties:
                    logger.info("Setting property %s: %s" % (property, image['meta'][property]))
                    self.conn.image.update_image(cloud_image.id, **{property: str(image['meta'][property])})

            logger.info("Checking status of '%s'" % name)
            if cloud_image.status != image['status'] and image['status'] == 'deactivated':
                logger.info("Deactivating image '%s'" % name)
                self.conn.image.deactivate_image(cloud_image.id)

            elif cloud_image.status != image['status'] and image['status'] == 'active':
                logger.info("Reactivating image '%s'" % name)
                self.conn.image.reactivate_image(cloud_image.id)

            logger.info("Checking visibility of '%s'" % name)
            if 'visibility' in versions[version]:
                visibility = versions[version]['visibility']
            else:
                visibility = image['visibility']

            if cloud_image.visibility != visibility:
                logger.info("Setting visibility of '%s' to '%s'" % (name, visibility))
                self.conn.image.update_image(cloud_image.id, visibility=visibility)

    def rename_images(self, name: str, sorted_versions: list, imported_image: Image, previous_image: Image) -> None:
        '''
        Rename outdated images in Glance (only applies to images of type multi)

        Params:
            name: the name of the image from images.yml
            sorted_versions: list with all sorted image versions
            imported_image: the newly imported image
            previous_image: the previous latest image
        '''
        cloud_images = self.get_images()

        if len(sorted_versions) > 1:
            latest = "%s (%s)" % (name, sorted_versions[-1])
            previous_latest = "%s (%s)" % (name, sorted_versions[-2])

            if name in cloud_images and previous_latest not in cloud_images:
                logger.info("Renaming %s to %s" % (name, previous_latest))
                self.conn.image.update_image(cloud_images[name].id, name=previous_latest)

            if latest in cloud_images:
                logger.info("Renaming %s to %s" % (latest, name))
                self.conn.image.update_image(cloud_images[latest].id, name=name)

        elif len(sorted_versions) == 1 and name in cloud_images:

            if previous_image['properties']['internal_version'] == 'latest':
                # if the last modification date cannot be found, use the creation date of the image instead
                create_date = str(datetime.strptime(previous_image.created_at, '%Y-%m-%dT%H:%M:%SZ').date())
                create_date = create_date.replace('-', '')

                previous_latest = "%s (%s)" % (name, create_date)

                logger.info('Setting internal_version: %s for %s' % (create_date, previous_latest))
                self.conn.image.update_image(previous_image.id, **{'internal_version': create_date})
            else:
                previous_latest = "%s (%s)" % (name, previous_image['properties']['internal_version'])

            logger.info("Renaming old latest '%s' to '%s'" % (name, previous_latest))
            self.conn.image.update_image(previous_image.id, name=previous_latest)

            logger.info("Renaming imported image '%s' to '%s'" % (imported_image.name, name))
            self.conn.image.update_image(imported_image.id, name=name)

        elif len(sorted_versions) == 1:
            latest = "%s (%s)" % (name, sorted_versions[-1])

            if latest in cloud_images:
                logger.info("Renaming %s to %s" % (latest, name))
                self.conn.image.update_image(cloud_images[latest].id, name=name)

    def manage_outdated_images(self, managed_images: set) -> list:
        '''
        Delete, hide or deactivate outdated images

        Params:
            managed_images: set of managed images
        Raises:
            Exception: when the image is still in use and cannot be deleted
            Exception: when the image cannot be deactivated or its visibility cannot be changed
        Returns:
            List with all images that are unmanaged and get affected by this method
        '''
        cloud_images = self.get_images()
        unmanaged_images = [x for x in cloud_images if x not in managed_images]

        for image in unmanaged_images:
            cloud_image = cloud_images[image]
            if self.CONF.delete and self.CONF.yes_i_really_know_what_i_do:
                try:
                    logger.info("Deactivating image '%s'" % image)
                    self.conn.image.deactivate_image(cloud_image.id)

                    logger.info("Setting visibility of '%s' to 'community'" % image)
                    self.conn.image.update_image(cloud_image.id, visibility='community')

                    logger.info("Deleting %s" % image)
                    self.conn.image.delete_image(cloud_image.id)
                except Exception as e:
                    logger.info("%s is still in use and cannot be deleted\n %s" % (image, e))

            else:
                logger.warning("Image %s should be deleted" % image)
                try:
                    if self.CONF.deactivate:
                        logger.info("Deactivating image '%s'" % image)
                        self.conn.image.deactivate_image(cloud_image.id)

                    if self.CONF.hide:
                        cloud_image = cloud_images[image]
                        logger.info("Setting visibility of '%s' to 'community'" % image)
                        self.conn.image.update_image(cloud_image.id, visibility='community')
                except Exception as e:
                    logger.error('An Exception occurred: \n%s' % e)
                    self.exit_with_error = True
        return unmanaged_images

    def check_image_metadata(self):
        """
        Retrieve metadata from managed images and check for compliance with SCS specifications:
        -> https://github.com/SovereignCloudStack/Docs/blob/main/Design-Docs/Image-Properties-Spec.md

        code adopted from https://github.com/SovereignCloudStack/Docs/blob/main/Design-Docs/tools/image-md-check.py
        """
        cloud_images = self.get_images()

        mand_images = ["Ubuntu 22.04", "Ubuntu 20.04", "Debian 11"]
        rec1_images = ["CentOS 8", "Rocky 8", "AlmaLinux 8", "Debian 10", "Fedora 36"]
        rec2_images = ["SLES 15SP4", "RHEL 9", "RHEL 8", "Windows Server 2022", "Windows Server 2019"]
        # sugg_images = ["openSUSE Leap 15.4", "Cirros 0.5.2", "Alpine", "Arch"]    # Ignore for now

        # properties inside of the main image dict
        image_properties = {
            "os_distro": {"mandatory": True, "values": None},
            "os_version": {"mandatory": True, "values": None},
            "architecture": {"mandatory": True, "values": ["x86_64", "aarch64", "risc-v"]},
            "hypervisor_type": {"mandatory": True, "values": ["qemu", "kvm", "xen", "hyper-v", "esxi", None]},
            "hw_rng_model": {"mandatory": False, "values": ["virtio", None]},
            "hw_disk_bus": {"mandatory": True, "values": ["virtio", "scsi", None]},
            "hash_algo": {"mandatory": False, "values": ["sha256", "sha512"]}
        }

        # properties inside of the image.properties dict
        custom_properties = {
            "image_build_date": {"mandatory": True, "values": None},
            "patchlevel": {"mandatory": False, "values": None},
            "image_original_user": {"mandatory": True, "values": None},
            "image_source": {"mandatory": True, "values": None},
            "image_description": {"mandatory": True, "values": None},
            "replace_frequency": {
                "mandatory": True,
                "values": ["yearly", "quarterly", "monthly", "weekly", "daily", "critical_bug", "never"]
            },
            "provided_until": {"mandatory": True, "values": None},
            "uuid_validity": {"mandatory": True, "values": None},
            "hotfix_hours": {"mandatory": False, "values": None}
        }

        for image_name in cloud_images.keys():
            try:
                image = self.conn.image.find_image(image_name)
            except DuplicateResource as exc:
                logger.error("Duplicate name: %s\n%s" % (image_name, str(exc)))
                self.exit_with_error = True
                continue
            if not image:
                logger.error("Image %s not found" % image_name)
                self.exit_with_error = True
                continue

            for prop in image_properties:
                if prop in image:
                    values = image_properties[prop]["values"]
                    if values and not image[prop] in values:
                        logger.error("Image %s: Value %s not allowed for property %s" % (image.name, image[prop], prop))
                        self.exit_with_error = True
                elif image_properties[prop]["mandatory"]:
                    logger.error("Image %s: Missing mandatory image property %s" % (image.name, prop))
                    self.exit_with_error = True
                else:
                    logger.warning("Image %s: Missing optional image property %s" % (image.name, prop))

            if ("hw_rng_model" in image and image.hw_rng_model == "scsi" and
                    "hw_disk_bus" in image and image.hw_disk_bus == "scsi"):
                if "hw_scsi_model" not in image:
                    logger.warning("Image %s: property hw_scsi_model is recommended when using 'scsi' for "
                                   "hw_rng_model and hw_disk_bus" % image.name)

            for prop in custom_properties:
                if prop in image.properties:
                    values = custom_properties[prop]["values"]
                    if values and not image.properties[prop] in values:
                        logger.error("Image %s: Value %s not allowed for property %s" %
                                     (image.name, image.properties[prop], prop))
                        self.exit_with_error = True
                elif custom_properties[prop]["mandatory"]:
                    logger.error("Image %s: Missing mandatory custom property %s" % (image.name, prop))
                    self.exit_with_error = True
                else:
                    logger.warning("Image %s: Missing optional custom property %s" % (image.name, prop))

            if "image_build_date" in image.properties:
                try:
                    time.strptime(image.properties["image_build_date"][:10], "%Y-%m-%d")
                except Exception:
                    logger.error("Image %s: invalid value for image_build_date %s" %
                                 (image.name, image.properties["build_date"]))
                    self.exit_with_error = True

            if "image_source" in image.properties:
                if not re.match(r"^(http|ftp)s?://\w*.*", image.properties["image_source"]):
                    logger.error("Image %s: image_source is not a valid URL" % image.name)
                    self.exit_with_error = True

            if "hotfix_hours" in image.properties:
                if type(image.properties["hotfix_hours"]) == str and not image.properties["hotfix_hours"].isdigit():
                    logger.warning("Image %s: custom property hotfix_hours is not numeric" % image.name)

            rec_image_name = "%s %s" % (image.os_distro, image.os_version)
            if not re.match(f"^{rec_image_name}", image.name, re.IGNORECASE):
                logger.warning("Image %s: Name does not start with recommended name %s" % (image.name, rec_image_name))

            rec_tag = "os:%s" % image.os_distro
            if rec_tag not in image.tags:
                logger.warning("Image %s: Missing recommended tag %s" % (image.name, rec_tag))
            if not any(tag.startswith("managed_by_") for tag in image.tags):
                logger.warning("Image %s: Missing recommended tag managed_by_" % image.name)

        # check for image completeness
        for image in mand_images:
            if image not in cloud_images:
                logger.warning("Mandatory image %s is missing" % image)
        for image in (*rec1_images, *rec2_images):
            if image not in cloud_images:
                logger.info("Recommended image %s is missing" % image)
        # Ignore sugg_images for now

    def validate_yaml_schema(self):
        """ Validate all image.yaml files against the SCS Metadata spec """
        schema = yamale.make_schema('etc/schema.yaml')
        try:
            for file in os.listdir(self.CONF.images):
                try:
                    data = yamale.make_data(self.CONF.images + file)
                    yamale.validate(schema, data)
                except YamaleError as e:
                    self.exit_with_error = True
                    for result in e.results:
                        logger.error("Error validating data '%s' with '%s'" % (result.data, result.schema))
                        for error in result.errors:
                            logger.error('\t%s' % error)
                else:
                    logger.debug("Image file %s is valid" % file)
        except FileNotFoundError:
            logger.error("Invalid path '%s'" % self.CONF.images)


def main():
    image_manager = ImageManager()
    typer.run(image_manager.create_cli_args)


if __name__ == '__main__':
    main()



import openstack
from loguru import logger
from munch import Munch
from unittest import TestCase

from src import manage

logger.remove()   # disable all logging from manage.py


class TestManageAPI(TestCase):

    def setUp(self):
        self.sot = manage.ImageManager()

        self.sot.CONF = Munch(
            latest=True,
            dry_run=False,
            use_os_hidden=False,
            delete=True,    # delete the image after the test
            yes_i_really_know_what_i_do=True,
            hide=False,
            deactivate=False,
            cloud='openstack',
            images='test/integration/fixtures/',
            name=None,
            tag='fake_tag',
            filter='',
            validate=False
        )
        self.image = self.sot.read_image_files()[0]
        self.assertEqual(self.image['name'], 'ubuntu-20.04')
        self.image['tags'].append(self.sot.CONF.tag)
        self.name = self.image['name'] + ' (1)'

    def test_api_functions(self):
        '''
        Test all used API functions, as they appear in src.manage.py
        Import the image, set its properties, rename it and delete it afterwards
        '''
        self.sot.conn = openstack.connect(cloud=self.sot.CONF.cloud)

        # make sure there are no images in the cloud already
        cloud_images = self.sot.get_images()
        self.assertEqual(cloud_images,
                         {},
                         f'Please make sure to delete any preexisting images with tag {self.sot.CONF.tag}')

        # test the image import
        imported_image = self.sot.import_image(self.image, self.name, self.image['versions'][0]['url'])
        self.assertEqual(imported_image.properties['os_glance_failed_import'], '')
        self.assertEqual(imported_image.visibility, 'private')
        self.assertEqual(imported_image.name, self.name)
        self.assertEqual(imported_image.min_ram, self.image['min_ram'])
        self.assertEqual(imported_image.min_disk, self.image['min_disk'])
        self.assertEqual(imported_image.tags, [self.sot.CONF.tag])
        self.assertEqual(imported_image.container_format, 'bare')
        self.assertEqual(imported_image.status, 'active')
        self.assertFalse(imported_image.is_hidden)

        # test set properties
        self.sot.set_properties(self.image, self.name, {'1': dict()}, '1', '', self.image['meta'])

        # assert the properties of the image got updated
        image = self.sot.get_images()[self.name]
        self.assertEqual(image.properties['os_glance_failed_import'], '')
        self.assertEqual(image.visibility, 'public')
        self.assertEqual(image.name, self.name)
        self.assertEqual(image.min_ram, self.image['min_ram'])
        self.assertEqual(image.min_disk, self.image['min_disk'])
        self.assertEqual(image.tags, [self.sot.CONF.tag])
        self.assertEqual(image.container_format, 'bare')
        self.assertEqual(image.status, 'active')
        self.assertEqual(image.architecture, self.image['meta']['architecture'])
        self.assertEqual(image.hw_disk_bus, self.image['meta']['hw_disk_bus'])
        self.assertEqual(image.hw_rng_model, self.image['meta']['hw_rng_model'])
        self.assertEqual(image.hw_scsi_model, self.image['meta']['hw_scsi_model'])
        self.assertEqual(image.hw_watchdog_action, self.image['meta']['hw_watchdog_action'])
        self.assertEqual(image.os_distro, self.image['meta']['os_distro'])
        self.assertEqual(image.os_version, self.image['meta']['os_version'])
        self.assertEqual(image.properties['image_original_user'], self.image['login'])
        self.assertEqual(image.properties['internal_version'], '1')
        self.assertFalse(image.is_hidden)

        # test image rename
        self.sot.rename_images(self.image['name'], ['1'], None, None)

        # assert the image got renamed
        cloud_images = self.sot.get_images()
        self.assertEqual(list(cloud_images.keys()), [self.image['name']])

        # Finally delete the image, also works as a cleanup
        res = self.sot.manage_outdated_images(set())
        self.assertEqual(res, [self.image['name']])

        # make sure the image is gone
        cloud_images = self.sot.get_images()
        self.assertEqual(cloud_images, {}, f"Cloud not delete image {self.image['name']}, please do so manually")



# NOTE(berendt): quick & dirty (but it works for the moment)
# source of latest URLs: https://gitlab.com/libosinfo/osinfo-db

from datetime import datetime
import os
import re
import shutil
import sys
import time
from urllib.parse import urlparse
from urllib.request import urlopen

from loguru import logger
from minio import Minio
from minio.error import S3Error
from munch import Munch
from natsort import natsorted
import patoolib
import requests
import ruamel.yaml
import typer
import yaml

app = typer.Typer()


def mirror_image(image, latest_url, CONF):
    client = Minio(
        CONF.minio_server,
        access_key=CONF.minio_access_key,
        secret_key=CONF.minio_secret_key,
    )

    version = image["versions"][0]
    version["source"] = latest_url

    path = urlparse(version["source"])
    dirname = image["shortname"]
    filename, fileextension = os.path.splitext(os.path.basename(path.path))

    if fileextension not in [".bz2", ".zip", ".xz"]:
        filename += fileextension

    shortname = image["shortname"]
    format = image["format"]
    new_version = version["version"]
    new_filename = f"{new_version}-{shortname}.{format}"

    try:
        client.stat_object(CONF.minio_bucket, os.path.join(dirname, new_filename))
        logger.info("'%s' available in '%s'" % (new_filename, dirname))
    except S3Error:
        logger.info("'%s' not yet available in '%s'" % (new_filename, dirname))
        logger.info("Downloading '%s'" % version["source"])
        response = requests.get(version["source"], stream=True)
        with open(os.path.basename(path.path), "wb") as fp:
            shutil.copyfileobj(response.raw, fp)
        del response

        if fileextension in [".bz2", ".zip", ".xz"]:
            logger.info("Decompressing '%s'" % os.path.basename(path.path))
            patoolib.extract_archive(os.path.basename(path.path), outdir=".")
            os.remove(os.path.basename(path.path))

        logger.info(
            "Uploading '%s' to '%s' as '%s'" % (filename, dirname, new_filename)
        )

        client.fput_object(
            CONF.minio_bucket, os.path.join(dirname, new_filename), filename
        )
        os.remove(filename)


def update_image(image, CONF):
    name = image["name"]
    logger.info(f"Checking image {name}")

    latest_url = image["latest_url"]
    logger.info(f"Latest download URL is {latest_url}")

    parsed_url = urlparse(latest_url)
    latest_filename = os.path.basename(parsed_url.path)

    latest_checksum_url = image["latest_checksum_url"]
    logger.info(f"Getting checksums from {latest_checksum_url}")

    result = requests.get(latest_checksum_url)
    checksums = {}

    checksum_type = "sha256"
    filename_pattern = None

    if image["shortname"] in ["centos-stream-8", "centos-stream-9", "centos-7"]:
        filename_pattern = latest_filename.replace("HEREBE", "")
        filename_pattern = filename_pattern.replace("DRAGONS", "")
    elif image["shortname"] in ["debian-10", "debian-11"]:
        checksum_type = "sha512"

    for line in result.text.split("\n"):
        if image["shortname"] == "rocky-9":
            splitted_line = re.split("\s+", line)  # noqa W605
            if splitted_line[0] == "SHA256":
                checksums[latest_filename] = splitted_line[3]
        elif image["shortname"] in [
            "ubuntu-14.04",
            "ubuntu-16.04",
            "ubuntu-16.04-minimal",
            "ubuntu-18.04",
            "ubuntu-18.04-minimal",
            "ubuntu-20.04",
            "ubuntu-20.04-minimal",
            "ubuntu-22.04",
            "ubuntu-22.04-minimal",
        ]:
            splitted_line = re.split("\s+", line)  # noqa W605
            if len(splitted_line) == 2:
                checksums[splitted_line[1][1:]] = splitted_line[0]
        elif image["shortname"] in ["centos-7"]:
            splitted_line = re.split("\s+", line)  # noqa W605
            if len(splitted_line) == 2:
                if re.search(filename_pattern, splitted_line[1]):
                    checksums[splitted_line[1]] = splitted_line[0]
        elif image["shortname"] in ["centos-stream-8", "centos-stream-9"]:
            splitted_line = re.split("\s+", line)  # noqa W605
            if splitted_line[0] == "SHA256" and re.search(
                filename_pattern, splitted_line[1][1:-1]
            ):
                checksums[splitted_line[1][1:-1]] = splitted_line[3]
        else:
            splitted_line = re.split("\s+", line)  # noqa W605
            if len(splitted_line) == 2:
                checksums[splitted_line[1]] = splitted_line[0]

    if filename_pattern:
        new_latest_filename = natsorted(checksums.keys())[-1]
        new_latest_url = latest_url.replace(latest_filename, new_latest_filename)

        logger.info(f"Latest URL is now {new_latest_url}")
        logger.info(f"Latest filename is now {new_latest_filename}")

        latest_filename = new_latest_filename
        latest_url = new_latest_url

    current_checksum = f"{checksum_type}:{checksums[latest_filename]}"
    logger.info(f"Checksum of current {latest_filename} is {current_checksum}")

    latest_version = image["versions"][0]
    latest_checksum = latest_version["checksum"]
    logger.info(f"Our checksum is {latest_checksum}")

    if latest_checksum != current_checksum:
        logger.info(f"Checking {latest_url}")

        conn = urlopen(latest_url, timeout=30)
        struct = time.strptime(
            conn.headers["last-modified"], "%a, %d %b %Y %H:%M:%S %Z"
        )
        dt = datetime.fromtimestamp(time.mktime(struct))

        new_version = dt.strftime("%Y%m%d")
        logger.info(f"New version is {new_version}")
        image["versions"][0]["version"] = new_version

        new_build_date = dt.strftime("%Y-%m-%d")
        logger.info(f"New build date is {new_build_date}")
        image["versions"][0]["build_date"] = new_build_date

        logger.info(f"New checksum is {current_checksum}")
        image["versions"][0]["checksum"] = current_checksum

        shortname = image["shortname"]
        format = image["format"]

        minio_server = str(CONF.minio_server)
        minio_bucket = str(CONF.minio_bucket)
        new_url = f"https://{minio_server}/{minio_bucket}/{shortname}/{new_version}-{shortname}.{format}"
        logger.info(f"New URL is {new_url}")
        image["versions"][0]["url"] = new_url

        mirror_image(image, latest_url, CONF)
        del image["versions"][0]["source"]

    else:
        logger.info(f"Image {name} is up-to-date, nothing to do")

    return image


@app.command()
def main(
    debug: bool = typer.Option(False, "--debug", help="Enable debug logging"),
    image: str = typer.Option("almalinux", help="Image to update"),
    minio_access_key: str = typer.Option(None, help="Minio access key"),
    minio_secret_key: str = typer.Option(None, help="Minio secret key"),
    minio_server: str = typer.Option("minio.services.osism.tech", help="Minio server"),
    minio_bucket: str = typer.Option("openstack-image-manager", help="Minio bucket"),
):

    CONF = Munch.fromDict(locals())

    if CONF.debug:
        level = "DEBUG"
    else:
        level = "INFO"

    logger.remove()  # remove the default sink
    log_fmt = (
        "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | "
        "<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"
    )
    logger.add(sys.stderr, format=log_fmt, level=level, colorize=True)

    p = f"etc/images/{image}.yml"

    with open(p) as fp:
        data = yaml.safe_load(fp)

    for index, image in enumerate(data["images"]):
        if "latest_url" in image:
            updated_image = update_image(image, CONF)
            data["images"][index] = updated_image

    with open(p, "w+") as fp:
        ryaml = ruamel.yaml.YAML()
        ryaml.explicit_start = True
        ryaml.indent(sequence=4, offset=2)
        ryaml.dump(data, fp)


if __name__ == "__main__":
    app()



__all__ = ['__version__']

import pbr.version

version_info = pbr.version.VersionInfo('openstack-image-manager')
# We have a circular import problem when we first run python setup.py sdist
# It's harmless, so deflect it.
try:
    __version__ = version_info.version_string()
except AttributeError:
    __version__ = None



import logging
import os
import typer

from munch import Munch
import openstack


logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
app = typer.Typer(add_completion=False)


@app.command()
def main(
        dry_run: bool = typer.Option(False, '--dry-run', help='Do not perform any changes'),
        action: str = typer.Option('add', help='Action'),
        cloud: str = typer.Option('service', help='Managed cloud'),
        image: str = typer.Option(..., help='Image to share'),
        project_domain: str = typer.Option('default', help='Target project domain'),
        target: str = typer.Option(..., help='Target project domain'),
        type: str = typer.Option('project', help='Project or domain')
):
    CONF = Munch.fromDict(locals())

    if "OS_AUTH_URL" in os.environ:
        conn = openstack.connect()
    else:
        conn = openstack.connect(cloud=CONF.cloud)
    image = conn.get_image(CONF.image)

    if CONF.type == "project":
        domain = conn.get_domain(name_or_id=CONF.project_domain)
        project = conn.get_project(CONF.target, domain_id=domain.id)

        if CONF.action == "add":
            share_image_with_project(CONF, conn, image, project)
        elif CONF.action == "del":
            unshare_image_with_project(CONF, conn, image, project)

    elif CONF.type == "domain":
        domain = conn.get_domain(name_or_id=CONF.target)
        projects = conn.list_projects(domain_id=domain.id)
        for project in projects:
            if CONF.action == "add":
                share_image_with_project(CONF, conn, image, project)
            elif CONF.action == "del":
                unshare_image_with_project(CONF, conn, image, project)


def unshare_image_with_project(CONF, conn, image, project):
    member = conn.image.find_member(project.id, image.id)

    if member:
        logging.info("del - %s - %s (%s)" % (image.name, project.name, project.domain_id))

        if not CONF.dry_run:
            conn.image.remove_member(member, image.id)


def share_image_with_project(CONF, conn, image, project):
    member = conn.image.find_member(project.id, image.id)

    if not member:
        logging.info("add - %s - %s (%s)" % (image.name, project.name, project.domain_id))

        if not CONF.dry_run:
            member = conn.image.add_member(image.id, member_id=project.id)

    if not CONF.dry_run and member.status != "accepted":
        logging.info("accept - %s - %s (%s)" % (image.name, project.name, project.domain_id))
        conn.image.update_member(member, image.id, status="accepted")


if __name__ == '__main__':
    app()



# NOTE(berendt): quick & dirty (but it works for the moment)

import logging
import os
import shutil
import patoolib
import requests
import yaml
import typer
from urllib.parse import urlparse

from minio import Minio
from minio.error import S3Error
from munch import Munch
from os import listdir
from os.path import isfile, join


app = typer.Typer(add_completion=False)


@app.command()
def main(
        debug: bool = typer.Option(False, '--debug', help='Enable debug logging'),
        dry_run: bool = typer.Option(False, '--dry-run', help='Do not perform any changes'),
        images: str = typer.Option('etc/images/', help='Path to the directory containing all image files'),
        minio_access_key: str = typer.Option(None, help='Minio access key'),
        minio_secret_key: str = typer.Option(None, help='Minio secret key'),
        minio_server: str = typer.Option('minio.services.osism.tech', help='Minio server'),
        minio_bucket: str = typer.Option('openstack-image-manager', help='Minio bucket')
):
    CONF = Munch.fromDict(locals())

    if CONF.debug:
        level = logging.DEBUG
        logging.getLogger("paramiko").setLevel(logging.DEBUG)
    else:
        level = logging.INFO
        logging.getLogger("paramiko").setLevel(logging.WARNING)
    logging.basicConfig(format='%(asctime)s - %(message)s', level=level, datefmt='%Y-%m-%d %H:%M:%S')

    onlyfiles = []
    for f in listdir(CONF.images):
        if isfile(join(CONF.images, f)):
            onlyfiles.append(f)

    all_images = []
    for file in onlyfiles:
        with open(join(CONF.images, file)) as fp:
            data = yaml.load(fp, Loader=yaml.SafeLoader)
            images = data.get('images')
            for image in images:
                all_images.append(image)

    client = Minio(
        CONF.minio_server,
        access_key=CONF.minio_access_key,
        secret_key=CONF.minio_secret_key
    )

    for image in all_images:
        for version in image['versions']:
            if 'source' not in version:
                continue

            logging.debug("source: %s" % version['source'])

            path = urlparse(version['source'])

            dirname = "%s/%s" % (image['shortname'], version['version'])
            filename, fileextension = os.path.splitext(os.path.basename(path.path))

            if fileextension not in ['.bz2', '.zip', '.xz']:
                filename += fileextension

            logging.debug("dirname: %s" % dirname)
            logging.debug("filename: %s" % filename)

            try:
                client.stat_object(CONF.minio_bucket, os.path.join(dirname, filename))
                logging.info("'%s' available in '%s'" % (filename, dirname))
            except S3Error:
                logging.info("'%s' not yet available in '%s'" % (filename, dirname))

                if not CONF.dry_run:
                    logging.info("Downloading '%s'" % version['source'])
                    response = requests.get(version['source'], stream=True)
                    with open(os.path.basename(path.path), 'wb') as fp:
                        shutil.copyfileobj(response.raw, fp)
                    del response

                    if fileextension in ['.bz2', '.zip', '.xz']:
                        logging.info("Decompressing '%s'" % os.path.basename(path.path))
                        patoolib.extract_archive(os.path.basename(path.path), outdir='.')
                        os.remove(os.path.basename(path.path))

                    logging.info("Uploading '%s' to '%s'" % (filename, dirname))

                    client.fput_object(CONF.minio_bucket, os.path.join(dirname, filename), filename)
                    os.remove(filename)


if __name__ == '__main__':
    app()



def setup(app):
    app.add_css_file("css/custom.css")


extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.todo",
    "sphinx_fontawesome",
]
source_suffix = ".rst"
master_doc = "index"
project = ""
copyright = "2022, OSISM GmbH"
author = "OSISM GmbH"
version = ""
release = ""
language = "en"
exclude_patterns = []
pygments_style = "sphinx"
todo_include_todos = True
html_theme = "sphinx_material"
html_show_sphinx = False
html_show_sourcelink = False
html_show_copyright = True
htmlhelp_basename = "openstack-image-manager"
html_theme_options = {
    "nav_title": "OSISM OpenStack Image Manager",
    "color_primary": "blue",
    "color_accent": "light-blue",
    "globaltoc_depth": 3,
    "globaltoc_collapse": True,
}
html_context = {}
html_title = "OSISM OpenStack Image Manager"
html_logo = "images/logo.png"
html_sidebars = {
    "**": ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]
}
latex_elements = {}
